﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Media.Media3D;
using System.Windows.Navigation;
using System.Windows.Shapes;
using HeuristicLab.Encodings.PackingEncoding.PackingPlan;
using HeuristicLab.Problems.BinPacking.Dimensions;
using HeuristicLab.Problems.BinPacking.PackingBin;
using HeuristicLab.Problems.BinPacking.PackingItem;

namespace HeuristicLab.Problems.BinPacking.Views.Shapes {
  public partial class Container3DView : UserControl {
    private Point startPos;
    private bool mouseDown = false;
    private double startAngleX;
    private double startAngleY;
    private int selectedItemIndex;

    public Container3DView() {
      InitializeComponent();
      Clear();
    }

    public void Clear() {
      meshMain.Positions.Clear();
      meshMain.TriangleIndices.Clear();

      mouseDown = false;
      startAngleX = 0;
      startAngleY = 0;
      selectedItemIndex = -1;

      AddCube(0, 0, 0, 1, 1, 1); // add an empty cube (TODO: remove)
    }


    public void ShowContainer(BinPacking<ThreeDimensionalPacking, CuboidPackingBin, CuboidPackingItem> packing) {
      var container = packing.BinMeasures;
      // TODO: draw transparent container
      // packingPlan3D.InitializeContainer(bin.Width, bin.Height, bin.Depth);

      foreach (var item in packing.ItemMeasures) {
        var position = packing.ItemPositions[item.Key];

        // ignore the item.Key
        // ignore the item.Material

        AddCube(position.X, position.Y, position.Z,
          position.Rotated ? item.Value.Depth : item.Value.Width,
          item.Value.Height,
          position.Rotated ? item.Value.Width : item.Value.Depth);
      }
    }

    public void SelectItem(int itemIndex) {
      // selection of an item should make all other items semi-transparent
      selectedItemIndex = itemIndex;
    }
    public void ClearSelection() {
      // remove all transparency
      selectedItemIndex = -1;
    }

    private void viewport3D1_MouseWheel(object sender, MouseWheelEventArgs e) {
      if (e.Delta > 0) {
        camMain.Position -= 0.5 * camMain.LookDirection;
      } else if (e.Delta < 0) {
        camMain.Position += 0.5 * camMain.LookDirection;
      }
    }

    private void viewport3D1_MouseMove(object sender, MouseEventArgs e) {
      if (!mouseDown) return;
      var pos = e.GetPosition((IInputElement)this);
      rotateX.Angle = startAngleX + (pos.X - startPos.X) / 4;
      rotateY.Angle = startAngleY + (pos.Y - startPos.Y) / 4;
    }

    private void viewport3D1_MouseDown(object sender, MouseButtonEventArgs e) {
      startAngleX = rotateX.Angle;
      startAngleY = rotateY.Angle;
      this.startPos = e.GetPosition((IInputElement)this);
      this.mouseDown = true;
    }

    private void viewport3D1_MouseUp(object sender, MouseButtonEventArgs e) {
      mouseDown = false;
    }


    #region helper for cubes

    private void AddCube(int x, int y, int z, int width, int height, int depth) {
      AddTriangles(AddPoints(x, y, z, width, height, depth));
    }

    private void AddTriangles(int[] pointIdx) {
      // back side
      meshMain.TriangleIndices.Add(pointIdx[0]);
      meshMain.TriangleIndices.Add(pointIdx[2]);
      meshMain.TriangleIndices.Add(pointIdx[1]);

      meshMain.TriangleIndices.Add(pointIdx[0]);
      meshMain.TriangleIndices.Add(pointIdx[3]);
      meshMain.TriangleIndices.Add(pointIdx[2]);

      // right side
      meshMain.TriangleIndices.Add(pointIdx[2]);
      meshMain.TriangleIndices.Add(pointIdx[6]);
      meshMain.TriangleIndices.Add(pointIdx[5]);

      meshMain.TriangleIndices.Add(pointIdx[1]);
      meshMain.TriangleIndices.Add(pointIdx[2]);
      meshMain.TriangleIndices.Add(pointIdx[5]);

      // left side
      meshMain.TriangleIndices.Add(pointIdx[7]);
      meshMain.TriangleIndices.Add(pointIdx[3]);
      meshMain.TriangleIndices.Add(pointIdx[4]);

      meshMain.TriangleIndices.Add(pointIdx[4]);
      meshMain.TriangleIndices.Add(pointIdx[3]);
      meshMain.TriangleIndices.Add(pointIdx[0]);

      // front side
      meshMain.TriangleIndices.Add(pointIdx[6]);
      meshMain.TriangleIndices.Add(pointIdx[7]);
      meshMain.TriangleIndices.Add(pointIdx[4]);

      meshMain.TriangleIndices.Add(pointIdx[6]);
      meshMain.TriangleIndices.Add(pointIdx[4]);
      meshMain.TriangleIndices.Add(pointIdx[5]);

      // top side
      meshMain.TriangleIndices.Add(pointIdx[3]);
      meshMain.TriangleIndices.Add(pointIdx[7]);
      meshMain.TriangleIndices.Add(pointIdx[6]);

      meshMain.TriangleIndices.Add(pointIdx[3]);
      meshMain.TriangleIndices.Add(pointIdx[6]);
      meshMain.TriangleIndices.Add(pointIdx[2]);

      // bottom side
      meshMain.TriangleIndices.Add(pointIdx[5]);
      meshMain.TriangleIndices.Add(pointIdx[4]);
      meshMain.TriangleIndices.Add(pointIdx[0]);

      meshMain.TriangleIndices.Add(pointIdx[5]);
      meshMain.TriangleIndices.Add(pointIdx[0]);
      meshMain.TriangleIndices.Add(pointIdx[1]);


    }

    private int[] AddPoints(int x, int y, int z, int w, int h, int d) {
      // ground
      meshMain.Positions.Add(new Point3D(x, y, z));
      meshMain.Positions.Add(new Point3D(x + w, y, z));
      meshMain.Positions.Add(new Point3D(x + w, y + h, z));
      meshMain.Positions.Add(new Point3D(x, y + h, z));
      // top
      meshMain.Positions.Add(new Point3D(x, y, z + d));
      meshMain.Positions.Add(new Point3D(x + w, y, z + d));
      meshMain.Positions.Add(new Point3D(x + w, y + h, z + d));
      meshMain.Positions.Add(new Point3D(x, y + h, z + d));

      return Enumerable.Range(meshMain.Positions.Count - 8, 8).ToArray();
    }
    #endregion
  }
}
